import re
import copy
import importlib as imp

from comsdk.graph import Graph, Func, State, Selector
from comsdk.edge import Edge


class Params():
    __slots__=(
        'module',
        'entry_func',
        'predicate',
        'selector',
        'function',
        'morphism',
        'parallelism',
        'comment',
        'order',
        'subgraph'
    )
    def __init__(self):
        for slot in self.__slots__:
            setattr(self, slot, None)
    
    def __str__(self):
        stri = ""
        for s in self.__slots__:
            stri += ((s+": {}, ".format(getattr(self, s))) if getattr(self, s) is not None else "")
        return stri

# entities = {}

class GraphFactory():
    __slots__ = (
        'name',
        'states', 
        'graph',
        'issub',
        'tocpp',
        'entities'
    )
    def __init__(self, tocpp=False):
        self.states = {}
        self.entities = {}
        self.tocpp = tocpp
        self.name = None
        self.issub = False

    def add_state(self, statename):
        if statename not in self.states:
            self.states[statename] = State(statename)
            if statename in self.entities:
                self.states[statename].comment = self.entities[statename].comment
    
    def _create_morphism(self, morphname=None):
        if morphname is None:
            return Func(), Func()
        pred_f, func_f = Func(), Func() 
        morph = self.entities[morphname]
        # print(morph)
        for m in morph.__slots__:
            if getattr(morph,m) != None:
                if m!="predicate" and m!="function":
                    raise Exception("ERROR: Morphisms could not have any params exept predicate and function!\n{}".format(morphname))
                if m=="predicate":
                    if getattr(morph,m) not in self.entities:
                        raise Exception("\tERROR: Predicate {} is not defined!".format(getattr(morph, m)))
                    pred = self.entities[getattr(morph, m)]
                    if self.tocpp:
                        pred_f = Func(pred.module, pred.entry_func, dummy=True)
                    else:
                        pred_f = Func(pred.module, pred.entry_func)
                if m=="function":
                    if getattr(morph,m) not in self.entities:
                       raise Exception("\tERROR: Function: {} is not defined!".format(getattr(morph, m)))
                    fu = self.entities[getattr(morph, m)]
                    if self.tocpp:
                        func_f = Func(fu.module, fu.entry_func, dummy=True)
                    else:
                        func_f = Func(fu.module, fu.entry_func)
        return pred_f, func_f


    def add_connection(self, st1, st2, morphism=None, ordr=0):
        pred, entr = self._create_morphism(morphism)
        self.states[st1].connect_to(self.states[st2], edge=Edge(pred, entr, order=ordr))
        print("{} -> {}".format(st1, st2))

    def build(self, nsub):
        print("BUILDING {}\nStates:".format(self.name))
        for s in self.states:
            print("\t"+ s)
        if self.issub:
            self.graph = Graph(self.states[self.name+str(nsub)+"_"+"__BEGIN__"], self.states[self.name+str(nsub)+"_"+"__END__"])
        else:    
            self.graph = Graph(self.states["__BEGIN__"], self.states["__END__"])
        self.graph.init_graph()
        for s in self.states:
            if s in self.entities and self.entities[s].selector is not None:
                selname = self.entities[s].selector
                if self.tocpp:
                    self.states[s].selector = Selector(len(self.states[s].transfers), self.entities[selname].module, self.entities[selname].entry_func, dummy=True)
                else:
                    self.states[s].selector = Selector(len(self.states[s].transfers), self.entities[selname].module, self.entities[selname].entry_func)
            else:
                self.states[s].selector =  Selector(len(self.states[s].transfers))
            if s in self.entities and self.entities[s].subgraph is not None:
                print("Replacing state {} with subgraph {}".format(s,self.entities[s].subgraph))
                parsr = Parser(subgraph=True)
                subgr = parsr.parse_file(self.entities[s].subgraph)
                self.states[s].replace_with_graph(subgr)
        return self.graph


class Parser():
    __slots__ = (
        'fact',
        'issub'
    )
    subgr_count = 0
    def __init__(self, tocpp=False, subgraph=False):
        self.fact = GraphFactory(tocpp=tocpp)
        self.fact.issub = subgraph
        self.issub = subgraph
        if subgraph:
            Parser.subgr_count+=1
    
    def _check_brackets(self, rawfile):
        br = 0
        qu = 0
        for char in rawfile:
            if char == "[":
                br+=1
            elif char == "{":
                br+=1
            elif char == "(":
                br+=1
            elif char == "]":
                br-=1
            elif char == "}":
                br-=1
            elif char == ")":
                br-=1
            elif char =="\"":
                qu+=1
        if br!=0 or qu%2!=0:
            raise Exception("Brackets or quotes do not match! Check your file")

    def _split_multiple(self,param):
        vals = {}
        first=True
        for s in param.__slots__:
            attr = getattr(param,s)
            if attr is not None and '\0' in attr:
                vals[s] = attr.split('\0')
        l=0
        for sl in vals:
            if l==0:
                l=len(vals[sl])
            elif l!=len(vals[sl]):
                raise Exception("\tERROR: Number of multiple params do not match", l)
        res = [copy.copy(param) for i in range(l)]
        for sl in vals:
            for i, _ in enumerate(res):
                setattr(res[i], sl, vals[sl][i])
        return res

    #Props is line "[proFp=smth, ...]"
    def _param_from_props(self,props):
        parm = Params()
        if props =="":
            return parm
        props = props.replace("]", '')
        if '(' in props:
            mchs = [m for m in re.finditer(r'\((\w+,)*\w+\)', props)]
            for m in mchs:
                props=props[:m.span()[0]]+(props[m.span()[0]:m.span()[1]]).replace(',','\0')+props[m.span()[1]:]
        if '\"' in props:
            mchs = [m for m in re.finditer(r'\".*\"', props)]
            for m in mchs:
                props=props[:m.span()[0]]+(props[m.span()[0]+1:m.span()[1]-1]).replace('\0',' ')+props[m.span()[1]:]
        props = props.replace("(","")
        props = props.replace(")","")
        rs =props.split(r",") #.split(r", ")
        for r in rs:
            r=r.split(r"=", 1)
            if r[0] in parm.__slots__:
                setattr(parm, r[0], r[1])
            else:
                raise Exception("\tERROR:Unknown parameter: "+ r[0])
        return parm

    def _param_from_entln(self, raw):
        res = re.split(r"\[", raw, 1)
        return res[0], self._param_from_props(res[1])

    def _multiple_morphs(self,props, n):
        p = self._param_from_props(props)
        if p.morphism is None:
            return  [copy.copy(p) for i in range(n)] 
        else:
            return self._split_multiple(p)

    def _topology(self,raw):
        spl = re.split(r"\s*(=>|->|\[|\])\s*", raw)
        spl = list(filter(lambda x: x!="[" and x!="]" and x!="", spl))
        left = spl[0].split(",")
        right = spl[2].split(",")
        if self.issub:
            for i in range(len(left)):
                left[i] = self.fact.name + str(Parser.subgr_count) + "_" + left[i]
            for i in range(len(right)):
                right[i] = self.fact.name + str(Parser.subgr_count) + "_" + right[i]
        if (len(left)>1) and (len(right)>1):
            raise Exception("ERROR: Ambigious multiple connection in line:\n\t{}".format(raw))
        # many to one conection
        elif len(left)>1:
            if len(spl) < 4:
                spl.append("")
            morphs = self._multiple_morphs(spl[3], len(left))
            if len(morphs)!=len(left):
                raise Exception("\tERROR: Count of edges do not match to count of states in many to one connection!\n\t\t{}".format(raw))
            self.fact.add_state(right[0])
            for i, st in enumerate(left):
                self.fact.add_state(st)
                self.fact.add_connection(st, right[0], morphs[i].morphism)
        # one to many connection, here could be selector
        elif len(right)>1:
            if len(spl) < 4:
                spl.append("")
            morphs = self._multiple_morphs(spl[3], len(right))
            self.fact.add_state(left[0])
            if len(morphs)!=len(right):
                raise Exception("\tERROR: Count of edges do not match to count of states in one to many connection!\n\t\t{}".format(raw))
            for i, st in enumerate(right):
                self.fact.add_state(st)
                self.fact.add_connection(left[0], st, morphs[i].morphism, morphs[i].order)
        # one to one connection
        else:
            self.fact.add_state(left[0])
            self.fact.add_state(right[0])
            if len(spl)==4:
                pr =self._param_from_props(spl[3])
                self.fact.add_connection(left[0], right[0], pr.morphism, ordr=pr.order if pr.order is not None else 0)
            elif len(spl)==3:
                self.fact.add_connection(left[0], right[0], None)

    def parse_file(self, filename):
        file = open(filename, "r")
        dot = file.read()
        self._check_brackets(dot)
        
        comments = [m for m in re.finditer(r'\".*\"', dot)]
        for m in comments:
            dot=dot[:m.span()[0]]+(dot[m.span()[0]:m.span()[1]]).replace(' ','\0')+dot[m.span()[1]:]
        dot = re.sub(r"[ \t\r]", "", dot) #deleting all spaces
        dot = re.sub(r"((digraph)|}|{)", "", dot)
        dot = re.sub(r"\/\/.*", "", dot)
        dot = re.sub(r"^\n$", "", dot)
        dotlines = dot.splitlines()
        dotlines = list(filter(None, dotlines))
        self.fact.name = dotlines[0]
        dotlines = dotlines[1:]
        # ent_re - regular expr for edges, states, functions properties
        ent_re = re.compile(r"^\w+\[.*\]$")
        # top_re - regular expr for topology properties, most time consuming one
        top_re = re.compile(r"^(\w+,?)+(->|=>)(\w+,?)+(\[(\w+=(\(?\w+,?\)?)+,?)+\])?")
        # (r"^\w[\w\s,]*(->|=>)\s*\w[\w\s,=\[\]()]*$")
        for i, ln in enumerate(dotlines):
            if ent_re.match(ln):
                name, parm = self._param_from_entln(ln)
                self.fact.entities[name] = parm
            elif top_re.match(ln):
                self._topology(ln)
        return self.fact.build(Parser.subgr_count)

    toload = []
    edgemap = {}

    # def _gen_branch(self,cur_state):
        # root = cur_state.name
        # edges = []
        # while cur_state is not None:
            # if len(cur_state.transfers)==0:
                # cur_state = None
            # elif len(cur_state.transfers)==1:
                # edges.append({"pred": str(cur_state.transfers[0].edge.pred_f),
                            #   "morph": str(cur_state.transfers[0].edge.morph_f)})
                # if str(cur_state.transfers[0].edge.pred_f) not in self.toload:
                    # self.toload.append(str(cur_state.transfers[0].edge))
                # if str(cur_state.transfers[0].edge.morph_f) not in self.toload:
                    # self.toload.append(str(cur_state.transfers[0].edge))
                # cur_state = cur_state.transfers[0].output_state
            # else:
                # pass
        # self.edgemap[root] = list(filter(lambda x: x["morph"] != '', edges))
        # return 


    checked=[]
    bushes = {}
    selectorends = {}
    def generate_cpp(self):
        self.fact.graph.init_state.input_edges_number =0
        states_to_check = [self.fact.graph.init_state]
        while len(states_to_check)!=0:
            for st in states_to_check:
                self.checked.append(st)
                states_to_check.remove(st)
                bush = _Bush(st)
                bush.grow_bush()
                self.bushes[st] = bush
                for outs in bush.outstates:
                    if outs not in states_to_check and outs not in self.checked:
                        states_to_check.append(outs)
        # for b in self.bushes:
            # print(b.name, "::")
            # for o in self.bushes[b].outstates:
                # print(o.name, len(o.transfers), end=' ')
            # print()
        # cur_b = self.bushes[self.fact.graph.init_state]
        # cur_b.selectorfin =  None
        # cur_b.token = 1
        # tocheck = cur_b.outstates
        # while len(tocheck)!=0:
            # for st in cur_b.outstates:
                # self.bushes[st].token = cur_b.level / len(cur_b.branches)
        # self.bushes[self.fact.graph.init_state].token
        send_token(self.fact.graph.init_state, 1, self.bushes, [])
        for b in self.bushes:
            print(b.name, "::")
            for inc in self.bushes[b].incomes:
                print(inc["st"].name, inc["i"], end =' ')
            print()
        
            # for br in bsh[b].branches:
                # if bsh[br.out].token-bsh[b] < 0:
                    # level+=1
        print("entities")
        for e in self.fact.entities:
            print(self.fact.entities[e])
        print_bush(self.fact.graph.init_state, self.fact.entities, self.bushes)

def print_bush(cur_state, entities, bushes):
    checked = []
    tocheck = [cur_state]
    while len(tocheck) !=0:
        cur_state=tocheck[0]
        cur_b = bushes[cur_state]
        # print("Name:", cur_state.name)
        if len(cur_b.branches)>1:
            print("{}:".format(cur_state.name))
        if len(cur_b.incomes)!=0:
            stri = "false "
            for inc in cur_b.incomes:
                stri += "|| SEL_{}[{}] ".format(inc["st"].name, inc["i"])
            print("if (!({}))".format(stri))
            print("{\n\tfor (int seli = 0; seli < 2; seli ++)\n", "\t\tSEL_{}[seli] =false;".format(cur_state.name),"\n}")
            if cur_state.name in entities and entities[cur_state.name].selector is not None:
                print("else \n{\n", "\tSEL_{} = {}();\n".format(cur_state.name, cur_state.selector), "\n}")
            else:
                print("else \n{\n\tfor (int seli = 0; seli < 2; seli ++)\n", "\t\tSEL_{}[seli] =true;".format(cur_state.name),"\n}")
        # if cur_state.name not in entities or entities[cur_state.name].selector is None:
            # for i,br in enumerate(cur_b.branches):
                # if br[len(br)-1].out not in tocheck:
                    # tocheck.append(br[len(br)-1].out)
                # print("append:", br[len(br)-1].out.name)
                # for b in br:
                    # if b.morph.name != "":
                    # 
                        # print("{}(data); {}(data);".format(b.pred, b.morph));
                # print("SEL_{}[{}] = true;".format(cur_state.name, i))
            # tocheck.remove(cur_state)
            # checked.append(cur_state)
            # print("remove: ",cur_state.name )
        # else:
        for i, br in enumerate(cur_b.branches):
            print("if (SEL_{}[{}])".format(cur_state.name, i))
            print("{")
            if br[len(br)-1].out not in tocheck:
                tocheck.append(br[len(br)-1].out)
            if br[len(br)-1].out in checked or br[len(br)-1].out is cur_state:
                for b in br:
                    if b.morph.name != "":
                        print("\t{}(data); {}(data);".format(b.pred, b.morph));
                print("\tgoto {};".format(br[len(br)-1].out.name))
                print("}")
                continue
            for b in br:
                if b.morph.name != "":
                    print("\t{}(data); {}(data);".format(b.pred, b.morph));
            print("}")
        tocheck.remove(cur_state)

                

    # if cur_state in checked:
        # return
    # else:
        # checked.append(cur_state)
    # if len(cur_b.outstates)==0:
        # return
    # checked.append(cur_state)
    # print(cur_b.state.name)
    # for i, br in enumerate(cur_b.branches):
        # if len(cur_b.branches)>1:
            # print("\t"*level,"if SEL_{}[{}]: {}".format(cur_state.name, i, len(cur_b.branches)))
        # for b in br:
            # print("\t"*(level+1), b.pred)
            # print("\t"*(level+1), b.morph)
        # out = cur_b.branches[i][len(cur_b.branches[i])-1].out
        # if out in checked:
            # print("\t"*level,"goto {}".format(out.name))
        # elif cur_b.token - bushes[cur_b.outstates[i]].token>0:
            # print("{} -> {} : {}".format(cur_state.name, out.name, cur_b.token - bushes[out].token))
            # level +=1
        # print_bush(out, level, bushes, checked)
        # if cur_b.token - bushes[out].token<0:
            # print("{} -> {} : {}".format(cur_state.name, out.name, cur_b.token - bushes[out].token))
            # level -=1


def send_token(cur_state, token, bushes, checked):
    cur_b = bushes[cur_state]
    if cur_state in checked:
        return
    cur_b.token += token
    # print(cur_b.state.name,len(cur_b.incomes), len(cur_b.outstates), cur_b.state.input_edges_number, cur_b.state.looped_edges_number)
    if len(cur_b.outstates)==0:
        return
    if len(cur_b.incomes) == cur_b.state.input_edges_number - cur_b.state.looped_edges_number:
        checked.append(cur_state)
        print("Sending:", cur_state.name)
        for i,br in enumerate(cur_b.branches):
            bushes[br[len(br)-1].out].incomes.append({"st":cur_state, "i":i})
            # print(bushes[br[len(br)-1].out].incomes)
            send_token(br[len(br)-1].out, cur_b.token / len(cur_b.branches), bushes, checked)

class _Tr():
    slots=(
        'pred',
        'morph',
        'out'
    )
    
    def __init__(self, transf):
        self.pred = transf.edge.pred_f
        self.morph = transf.edge.morph_f
        self.out = transf.output_state

    def __str__(self):
        return "{}\n{}\n".format(self.pred, self.morph)
            

class _Bush():
    __slots__=(
        'state', 
        'selector',
        'branches',
        'outstates',
        'token',
        'incomes',
        'selectorfin'
    )

    def __init__(self, state):
        self.state = state
        self.selector = state.selector
        self.branches = []
        self.outstates = []
        self.token = 0
        self.incomes = []

    def grow_bush(self):
        for t in self.state.transfers:
            branch = [_Tr(t)]
            self._gen_branch(t.output_state, branch)
        
    def _gen_branch(self, cur_state, branch):
        # print("generating", cur_state.name)
        while len(cur_state.transfers)==1 and cur_state.input_edges_number==1:
            tr = cur_state.transfers[0]
            branch.append(_Tr(tr))
            cur_state = tr.output_state
        self.branches.append(branch)
        if cur_state not in self.outstates:
            self.outstates.append(cur_state)
   

